/*
 * This file and everything in this package is covered under
 * the OPENdj Software License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.opendj
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-chat-server/src/org/opendj/server/chat/ChatConnection.java,v 1.1.1.1 2001/09/29 20:21:13 jonathan Exp $
 * $Log: ChatConnection.java,v $
 * Revision 1.1.1.1  2001/09/29 20:21:13  jonathan
 * Imported opendj-chat-server
 *
 */
package org.opendj.server.chat;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.Socket;
import java.net.InetAddress;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import org.opendj.beans.channel.ChannelBean;
import org.opendj.beans.genre.GenreBean;
import org.opendj.beans.mood.MoodBean;
import org.opendj.beans.dj.DjBean;
import org.opendj.beans.dj.NoSuchUserException;
import org.opendj.beans.dj.BadPasswordException;
import org.opendj.util.Util;
import org.opendj.util.DBUtil;
import org.opendj.util.ChannelTime;
import org.opendj.util.debug;

/**
 */
public class ChatConnection implements Runnable {

    private ChatServer  itsServer      = null;
    private Socket      itsSocket      = null;
    private ChatSocket  itsChatSocket  = null;
    private PStore      DB             = null;

    private long itsLoginTime = 0;
    private long itsID = -1;
    
    private ChatRoom itsRoom = null;
    private String itsRoomName = null;
    private String itsUserName = null;
    private DjBean itsDjBean = null;

    // Connection states 
    private static final int STATE_INIT         = 0;
    private static final int STATE_INROOM       = 1;
    private static final int STATE_NAMED        = 2;
    private static final int STATE_CLOSED       = 3;

    private int itsState = STATE_INIT;

    /**
     * Create a new ChatConnection
     *
     * @param theServer the ChatServer that accepted the socket connection
     * @param theSocket the socket connection to the remote computer
     * @param theID a unique identifier for this connection
     * @exception IOException if there is a problem setting up the connection
     */
    public ChatConnection ( ChatServer theServer, Socket theSocket, long theID ) 
        throws IOException {

        itsServer = theServer;
        itsSocket = theSocket;
        itsID = theID;
        itsLoginTime = System.currentTimeMillis();
        itsState = STATE_INIT;

        DB = itsServer.getPStore();

        log("Created connection for remote IP " + itsSocket.getInetAddress());
        itsChatSocket = new ChatSocket(this, itsSocket);
        itsRoomName = null;
        itsUserName = null;
    }

    public void run () {

        // Should we even be allowed to connect?
        try {
            if ( DB.isBlockedIP ( getIPAddress() ) ) {
                bootSelf();
                return;
            }
        } catch ( Exception e ) {
            log("run: Error checking if IP is blocked: " 
                + e.toString() + "\n" + Util.getStackTrace(e));
            bootSelf();
            return;
        }
        
        // Hand control over to ChatSocket
        try {
            itsChatSocket.run();
        } catch ( Exception e ) {
            log("run: " + e.toString() + "\n" + Util.getStackTrace(e));
        } finally {
            close();
        }
    }

    public void pong () {
        // Do nothing, all we need is that 
        // this keeps the Reader socket from timing out.
    }

    public void hello ( String theChatRoomName ) {
        if ( itsState != STATE_INIT ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        itsRoom = itsServer.getRoom( theChatRoomName );
        if ( itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_NOSUCHROOM);
            return;
        } 

        int aStatus = itsRoom.addChatter(this);
        switch ( aStatus ) {
        case StatusCode.HELLO_OK:
            log("Client successfully joined room '" + theChatRoomName + "'");
            itsState = STATE_INROOM;
            break;

        case StatusCode.HELLO_ROOMFULL:
            itsChatSocket.sendData(Protocol.RES_ROOMFULL);
            break;

        case StatusCode.HELLO_BLOCKED:
            close();
            break;

        default:
            log("join: ChatRoom.addChatter returned unknown StatusCode: " + aStatus);
            itsChatSocket.sendData(Protocol.RES_SERVER_ERROR);
        }
    }

    public void join ( String theUserName ) {
        if ( itsState != STATE_INROOM || itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        int aStatus = itsRoom.nameChatter( this, theUserName );
        switch ( aStatus ) {
        case StatusCode.NAME_OK:
            itsUserName = theUserName;
            itsState = STATE_NAMED;
            itsChatSocket.sendData(Protocol.RES_NAMEOK, itsUserName);
            break;

        case StatusCode.NAME_MUSTBEDJ:
            itsChatSocket.sendData(Protocol.RES_MUSTBEDJ);
            break;

        case StatusCode.NAME_TOOSHORT:
            itsChatSocket.sendData(Protocol.RES_NAMETOOSHORT);
            break;

        case StatusCode.NAME_TOOLONG:
            itsChatSocket.sendData(Protocol.RES_NAMETOOLONG);
            break;

        case StatusCode.NAME_TAKEN:
            itsChatSocket.sendData(Protocol.RES_NAMETAKEN);
            break;

        case StatusCode.NAME_ISDJNAME:
            itsChatSocket.sendData(Protocol.RES_NAMEISDJ);
            break;

        case StatusCode.NAME_BLOCKED:
            close();
            break;

        default:
            log("join: ChatRoom.nameChatter returned unknown StatusCode: " + aStatus);
            itsChatSocket.sendData(Protocol.RES_SERVER_ERROR);
        }
    }

    public void joinAsDj ( String theEmail, String thePassword ) {
        if ( itsState != STATE_INROOM || itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        joinAsDjWithEncryptedPassword( theEmail, 
                                       Util.getMD5Base64(thePassword) );
    }

    public void joinAsDjWithEncryptedPassword ( String theEmail, String thePassword ) {
        if ( itsState != STATE_INROOM || itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        log("joinAsDjWithEncryptedPassword('" + theEmail + "', '" + thePassword + "')");
        itsDjBean = new DjBean();

        int aStatus = -1;
        try {
            log("trying loginDj('" + theEmail + "', '" + thePassword + "'");
            itsDjBean = DB.loginDj( theEmail, thePassword );

        } catch ( NoSuchUserException nsue ) {
            itsDjBean = null;
            itsChatSocket.sendData(Protocol.RES_NOSUCHDJ);
            return;

        } catch ( BadPasswordException bpe ) {
            itsDjBean = null;
            itsChatSocket.sendData(Protocol.RES_BADPASSWORD);
            return;

        } catch ( Exception e ) {
            log("PStore.loginDj failed: " + e.toString() + "\n" + Util.getStackTrace(e));
            itsDjBean = null;
            itsChatSocket.sendData(Protocol.RES_SERVER_ERROR);
            return;
        }

        log("calling nameDJChatter('" + itsDjBean.getdjname() + "'");
        aStatus = itsRoom.nameDJChatter( this, itsDjBean.getdjname() );
        switch ( aStatus ) {
        case StatusCode.NAME_OK:
            itsUserName = itsDjBean.getdjname();
            itsState = STATE_NAMED;
            itsChatSocket.sendData(Protocol.RES_NAMEOK, itsUserName);
            break;

        case StatusCode.NAME_TAKEN:
            itsDjBean = null;
            itsChatSocket.sendData(Protocol.RES_NAMETAKEN);
            break;

        case StatusCode.NAME_BLOCKED:
            close();
            break;

        default:
            itsDjBean = null;
            log("join: ChatRoom.nameDJChatter returned unknown StatusCode: " + aStatus);
            itsChatSocket.sendData(Protocol.RES_SERVER_ERROR);
        }
    }

    public void message ( String theMessage, String theTarget ) {
        if ( itsState != STATE_NAMED || itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        itsRoom.sendMessage( this, theMessage, theTarget );
    }

    public void listusers () {
        if ( !(itsState == STATE_INROOM || itsState == STATE_NAMED) || itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        List someUsers = itsRoom.getUserNames();
        Iterator i = someUsers.iterator();
        while ( i.hasNext() ) {
            itsChatSocket.sendData(Protocol.RES_LUSER, i.next().toString());
        }
    }

    public void end () {
        if ( itsRoom != null ) {
            itsRoom.removeChatter(this, StatusCode.LEAVE_USER);
        }
        close();
    }

    public void boot ( String theUserToBoot ) {
        if ( itsState != STATE_NAMED || itsRoom == null ) {
            itsChatSocket.sendData(Protocol.RES_ILLEGAL_STATE);
            return;
        }
        int aStatus = itsRoom.bootUser(this, theUserToBoot);
        switch ( aStatus ) {
        case StatusCode.BOOT_OK:
            return;
        case StatusCode.BOOT_NOTMODERATOR:
            itsChatSocket.sendData(Protocol.RES_NOTMODERATOR);
            break;
        case StatusCode.BOOT_CANNOTBOOTSELF:
            itsChatSocket.sendData(Protocol.RES_CANTBOOTSELF);
            break;
        case StatusCode.BOOT_NOSUCHUSER:
            itsChatSocket.sendData(Protocol.RES_NOSUCHUSER);
            break;
        default:
            log("Unknown StatusCode returned from ChatRoom.bootUser: " + aStatus);
            itsChatSocket.sendData(Protocol.RES_SERVER_ERROR);
            return;
        }
    }

    public void sendMessage ( ChatConnection theSource, String theMessage ) {
        String aUserName = theSource.getUserName();
        if ( Util.nz(aUserName).length() == 0 ) {
            log("sendMessage: no user specified");
        } else {
            itsChatSocket.sendData(Protocol.RES_MESSAGE, theMessage, aUserName);
        }
    }

    public void sendPrivateMessage ( ChatConnection theSource, String theMessage ) {
        String aUserName = theSource.getUserName();
        if ( Util.nz(aUserName).length() == 0 ) {
            log("sendPrivateMessage: no user specified");
        } else {
            itsChatSocket.sendData(Protocol.RES_PRIVATEMESSAGE, theMessage, aUserName);
        }
    }

    public void sendJoinMessage ( String theName ) {
        itsChatSocket.sendData(Protocol.RES_JOIN, theName);
    }

    public void sendLeaveMessage ( String theName, int theReason ) {
        switch ( theReason ) {
        case StatusCode.LEAVE_USER:
            itsChatSocket.sendData(Protocol.RES_LEAVE, theName);
            break;
        case StatusCode.LEAVE_TIMEDOUT:
            itsChatSocket.sendData(Protocol.RES_LEAVE_TIMEDOUT, theName);
            break;
        case StatusCode.LEAVE_BOOTED:
            itsChatSocket.sendData(Protocol.RES_LEAVE_BOOTED, theName);
            break;
        default:
            log("sendLeaveMessage: unknown reason: " + theReason);
            itsChatSocket.sendData(Protocol.RES_LEAVE, theName);
            break;
        }
    }

    public void bootSelf () {
        close();
    }

    /**
     * Closes the connection and all associated resources.
     */
    public synchronized void close () {
        // log("ChatConnection.close called from: " + Util.getStackTrace(new Exception()));
        if ( itsChatSocket != null ) {
            itsChatSocket.finish();
            itsChatSocket = null;
        }
        finish();
    }

    public void finish () {
        // log("ChatConnection.finish called from: " + Util.getStackTrace(new Exception()));
        if ( itsSocket != null ) {
            try { itsSocket.close(); } catch ( Exception e ) {}
            itsSocket = null;
            log("Connection closed");
        }
        if ( itsRoom != null ) {
            itsRoom.removeChatter(this, StatusCode.LEAVE_USER);
            itsRoom = null;
        }
    }

    public DjBean getDjBean () { return itsDjBean; }

    /**
     * Returns true if this connection is still OK
     *
     * @return true if this connection is still in use
     */
    public boolean isAlive () { 
        return (itsSocket != null);
    }

    public int getState () { return itsState; }

    public String getUserName () { return itsUserName; }

    public String getIPAddress() { return itsSocket.getInetAddress().getHostAddress(); }

    /**
     * Logs some data to the server
     *
     * @param s the string to log
     */
    public void log ( String s ) {
        itsServer.log("[chc-" + itsID + "] " + s);
    }
}
