/*
 * This file and everything in this package is covered under
 * the OPENdj Software License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.opendj
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-chat-server/src/org/opendj/server/chat/ChatRoom.java,v 1.1.1.1 2001/09/29 20:21:13 jonathan Exp $
 * $Log: ChatRoom.java,v $
 * Revision 1.1.1.1  2001/09/29 20:21:13  jonathan
 * Imported opendj-chat-server
 *
 */
package org.opendj.server.chat;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Iterator;
import java.util.Vector;
import org.opendj.beans.dj.DjBean;
import org.opendj.util.Util;

public class ChatRoom {

    ChatServer itsServer = null;

    // DB fields
    private PStore  DB               = null;
    private String  itsRoomname      = null;
    private int     itsMaxchatters   = -1;
    private boolean itsIspublic      = false;
    private int     itsModeratordjid = -1;
    private String  itsDescription   = null;

    // Runtime status stuff
    List itsChatters = null;
    Map itsChattersByName = null;
    Map itsChattersByConnection = null;

    public ChatRoom (ChatServer theServer) {
        itsServer = theServer;
        DB = theServer.getPStore();
        itsChatters = new Vector();
        itsChattersByName = new HashMap();
        itsChattersByConnection = new HashMap();
    }

    public String getRoomname() { return itsRoomname; }
    public void setRoomname ( String theRoomname ) { itsRoomname = theRoomname; }

    public int getMaxchatters() { return itsMaxchatters; }
    public void setMaxchatters ( int theMaxchatters ) { itsMaxchatters = theMaxchatters; }

    public boolean getIspublic() { return itsIspublic; }
    public void setIspublic ( boolean theIspublic ) { itsIspublic = theIspublic; }

    public int getModeratordjid() { return itsModeratordjid; }
    public void setModeratordjid ( int theModeratordjid ) { itsModeratordjid = theModeratordjid; }

    public String getDescription() { return itsDescription; }
    public void setDescription ( String theDescription ) { itsDescription = theDescription; }

    public void init ( ResultSet theRS ) throws SQLException {
        setRoomname     (theRS.getString("roomname"));
        setMaxchatters  (theRS.getInt   ("maxchatters"));
        setIspublic     (theRS.getInt   ("ispublic") == 1);
        setModeratordjid(theRS.getInt   ("moderatordjid"));
        setDescription  (theRS.getString("description"));
    }

    public synchronized int addChatter ( ChatConnection theConn ) {
        if ( itsChatters.size() >= getMaxchatters() ) {
            return StatusCode.HELLO_ROOMFULL;
        }
        // Make sure they are not a booted user
        try {
            if ( DB.isBlockedIP( theConn.getIPAddress() ) ) {
                return StatusCode.HELLO_BLOCKED;
            }
        } catch ( Exception e ) {
            log("addChatter: error checking for blocked IP: "
                + e.toString() + "\n" + Util.getStackTrace(e));
            return StatusCode.HELLO_ERROR;
        }
        itsChatters.add(theConn);
        return StatusCode.HELLO_OK;
    }

    public synchronized void removeChatter( ChatConnection theConn ) {
        removeChatter(theConn, StatusCode.LEAVE_USER);
    }

    public synchronized void removeChatter( ChatConnection theConn, int theReason ) {
        itsChatters.remove(theConn);
        String aChatterName = (String) itsChattersByConnection.get(theConn);
        itsChattersByConnection.remove(theConn);
        if ( aChatterName != null ) {
            itsChattersByName.remove(aChatterName);
            notifyLeave(aChatterName, theReason);
        }
    }

    public synchronized int bootUser ( ChatConnection theConn, String theUserToBoot ) {

        ChatConnection aUserToBootConn = null;

        // Get the dj associated with the requestor
        DjBean aDjBean = theConn.getDjBean();
        if ( aDjBean == null ) return StatusCode.BOOT_NOTMODERATOR;

        // Find the user
        aUserToBootConn = (ChatConnection) itsChattersByName.get(theUserToBoot);
        if ( aUserToBootConn == null ) {
            return StatusCode.BOOT_NOSUCHUSER;
        }

        // Cannot boot self
        if ( theUserToBoot.equals(theConn.getUserName()) ) {
            return StatusCode.BOOT_CANNOTBOOTSELF;
        }

        // Refresh our internal state to reflect any updated moderatorid
        try {
            DB.refreshChatRoom(this);
        } catch ( Exception e ) {
            log("Could not refresh chat room: " 
                + e.toString() + "\n" + Util.getStackTrace(e));
            return StatusCode.BOOT_ERROR;
        }

        if ( aDjBean.getdjid() != itsModeratordjid ) {
            return StatusCode.BOOT_NOTMODERATOR;
        }

        // Send the connection a "boot" message.  Forcibly remove them here,
        // even though the connection itself should try to remove itself as well
        removeChatter(aUserToBootConn, StatusCode.LEAVE_BOOTED);
        try {
            DB.addBootRecord(itsModeratordjid, 
                             theUserToBoot, 
                             aUserToBootConn.getIPAddress());
        } catch ( Exception e ) {
            log("Could not create chat boot record: " 
                + e.toString() + "\n" + Util.getStackTrace(e));
            return StatusCode.BOOT_ERROR;
        }
        aUserToBootConn.bootSelf();
        return StatusCode.BOOT_OK;
    }

    public synchronized int nameChatter ( ChatConnection theConn, String theUserName ) {

        if ( !itsIspublic ) return StatusCode.NAME_MUSTBEDJ;

        if ( theUserName == null || theUserName.length() < 4 ) {
            return StatusCode.NAME_TOOSHORT;
        }
        if ( theUserName.length() > 100 ) {
            return StatusCode.NAME_TOOLONG;
        }

        // First check local names
        ChatConnection aConn = (ChatConnection) itsChattersByName.get(theUserName);
        if ( aConn == null ) {
            // Are they a blocked user?
            try {
                if ( DB.isBlockedName(theUserName) ) {
                    // Block this IP address too...
                    try {
                        DB.addBootRecord( 0, theUserName, theConn.getIPAddress() );
                    } catch ( Exception e ) {
                        log("nameChatter: error adding new boot record: " 
                            + e.toString() + "\n" + Util.getStackTrace(e));
                    }
                    return StatusCode.NAME_BLOCKED;
                }
            } catch ( Exception e ) {
                log("nameChatter: error checking for blocked user: "
                    + e.toString() + "\n" + Util.getStackTrace(e));
                return StatusCode.NAME_ERROR;
            }

            // Check all DJ names...
            try {
                if ( DB.checkDJNameExists(theUserName) ) {
                    return StatusCode.NAME_ISDJNAME;
                }
            } catch ( Exception e ) {
                log("PStore.checkDJNameExists: " + e.toString() 
                    + "\n" + Util.getStackTrace(e));
                return StatusCode.NAME_TAKEN; // just to be safe
            }

            itsChattersByName.put( theUserName, theConn );
            itsChattersByConnection.put( theConn, theUserName );

            // Let everyone know about the new person
            notifyJoin(theUserName);
            return StatusCode.NAME_OK;

        } else {
            return StatusCode.NAME_TAKEN;
        }
    }

    public synchronized int nameDJChatter ( ChatConnection theConn, 
                                            String theDJName ) {

        // First check local names
        ChatConnection aConn = (ChatConnection) itsChattersByName.get(theDJName);
        if ( aConn == null ) {
            // Are they a blocked user?
            try {
                if ( DB.isBlockedName(theDJName) ) {
                    // Block this IP address too...
                    try {
                        DB.addBootRecord( 0, theDJName, theConn.getIPAddress() );
                    } catch ( Exception e ) {
                        log("nameChatter: error adding new boot record: " 
                            + e.toString() + "\n" + Util.getStackTrace(e));
                    }
                    return StatusCode.NAME_BLOCKED;
                }
            } catch ( Exception e ) {
                log("nameDJChatter: error checking for blocked user: "
                    + e.toString() + "\n" + Util.getStackTrace(e));
                return StatusCode.NAME_ERROR;
            }

            itsChattersByName.put( theDJName, theConn );
            itsChattersByConnection.put( theConn, theDJName );

            // Let everyone know about the new person
            notifyJoin(theDJName);

            return StatusCode.NAME_OK;
        } else {
            return StatusCode.NAME_TAKEN;
        }
    }

    public synchronized void sendMessage ( ChatConnection theConn, 
                                           String theMessage,
                                           String theTarget ) {
        // If the target is null or empty, send to everyone
        ChatConnection aConn = null;
        if ( Util.nz(theTarget).length() == 0 ) {
            Iterator i = itsChatters.iterator();
            while ( i.hasNext() ) {
                aConn = (ChatConnection) i.next();
                aConn.sendMessage(theConn, theMessage);
            }
            try {
                DB.recordMessage( this, theConn, theMessage );
            } catch ( Exception e ) {
                log("sendMessage: error recording message: " 
                    + e.toString() + "\n" + Util.getStackTrace(e));
            }

        } else {
            // Only send to theTarget
            aConn = (ChatConnection) itsChattersByName.get(theTarget);
            if ( aConn == null ) {
                theConn.log("ChatRoom.sendMessage: target '" + theTarget + "' does not exist");
                return;
            }
            aConn.sendPrivateMessage(theConn, theMessage);
        }
    }

    public synchronized List getUserNames () {
        return new Vector(itsChattersByName.keySet());
    }

    private void notifyJoin ( String theName ) {
        Iterator i = itsChatters.iterator();
        ChatConnection aConn = null;
        while ( i.hasNext() ) {
            aConn = (ChatConnection) i.next();
            log("notifyJoin: sending join message to '" + aConn.getUserName() + "'");
            aConn.sendJoinMessage(theName);
        }
    }

    private void notifyLeave ( String theName, int theReason ) {
        Iterator i = itsChatters.iterator();
        ChatConnection aConn = null;
        while ( i.hasNext() ) {
            aConn = (ChatConnection) i.next();
            log("notifyLeave: sending leave message to '" + aConn.getUserName() + "'");
            aConn.sendLeaveMessage(theName, theReason);
        }
    }

    private void log ( Exception e ) {
        log(Util.getStackTrace(e));
    }

    private void log ( String s ) {
        itsServer.log("[ChatRoom {" + itsRoomname + "}] " + s);
    }
}
