/*
 * This file and everything in this package is covered under
 * the OPENdj Software License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.opendj
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-chat-server/src/org/opendj/server/chat/ChatServer.java,v 1.1.1.1 2001/09/29 20:21:13 jonathan Exp $
 * $Log: ChatServer.java,v $
 * Revision 1.1.1.1  2001/09/29 20:21:13  jonathan
 * Imported opendj-chat-server
 *
 */
package org.opendj.server.chat;

import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.FileInputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.sql.DriverManager;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;
import org.opendj.util.Util;
import org.opendj.util.Config;
import org.opendj.util.debug;
import org.opendj.util.service.Service;
import org.opendj.util.service.ServiceException;

public class ChatServer implements Service {

    public static final int    DEFAULT_PORT          = 8029;
    public static final String THREAD_GROUP_NAME     = "ChatServer-ThreadGroup";
    public static final long   LOOP_DELAY_SLEEP_MS   = 100;
    public static final long   LOOP_CLEANUP_COUNT    = 10;
    public static final int    ACCEPT_TIMEOUT_MILLIS = 1000;

    // Property variables -- initialized with the init method
    private int        itsPort        = DEFAULT_PORT;
    private Properties itsProps       = null;

    // Socket variables
    private ServerSocket itsServerSocket = null;

    private ThreadGroup  itsThreadGroup = null;
    private List         itsConnections = null;
    private long         itsCurrentID = 0;

    // DB vars
    private String itsDBDriver   = null;
    private String itsDBURL      = null;
    private String itsDBUser     = null;
    private String itsDBPassword = null;
    private PStore DB            = null;

    private Map    itsChatRooms  = null;

    private volatile boolean isDone = false;
    private volatile Thread itsMainThread = null;

    /**
     * Main method for the ChatServer.
     * Does basic validation on command line arguments
     *
     * @param args the command line arguments.
     */
    public static void main ( String[] args ) {

        ChatServer aServer = new ChatServer();

        try {
            aServer.init(Config.global().getProperties());
        } catch ( Exception e ) {
            System.err.println("ChatServer.main: Error initializing ChatServer: " 
                               + e.toString() + "\n" + Util.getStackTrace(e));
            return;
        }

        try {
            aServer.run();
        } catch ( Exception e ) {
            System.err.println("Error running ChatServer: " + e.toString());
            Util.getStackTrace(e);
            return;
        }
    }

    public void initService () throws ServiceException {
        try {
            Config aConfig = Config.global();
            init(aConfig.getProperties());
        } catch ( Exception e ) {
            throw new ServiceException(e);
        }
    }

    public void stopService () {
        isDone = true;
        log("shutting down...");
        if ( itsMainThread != null ) {
            try { itsServerSocket.close(); } catch ( Exception e ) {}
            try { itsMainThread.interrupt(); } catch ( Exception e ) {}
            try { itsMainThread.join(2000); } catch ( Exception e ) {}
            if ( itsMainThread.isAlive() ) {
                log("main thread did not exit gracefully, terminating with Thread.stop");
                itsMainThread.stop();
                cleanup();
                log("halted due to Thread.stop");
            }
        }
    }

    /**
     * Default constructor.
     */
    public ChatServer() {}

    /**
     * Performs initializations.  Sets up server socket and persistent store.
     *
     * @param theProps the properties to the server as specified in a properties file.
     * @exception ChatException if an exception occurs while initializing
     */
    public void init ( Properties theProps ) throws ChatException {

        itsProps = theProps;

        try {
            itsPort = Integer.parseInt(theProps.getProperty(Config.PROP_CHATSERVER_PORT));
        } catch ( Exception e ) {
            itsPort = DEFAULT_PORT;
        }

        try {
            itsServerSocket = new ServerSocket( itsPort );
            itsServerSocket.setSoTimeout(ACCEPT_TIMEOUT_MILLIS);
        } catch ( Exception e ) {
            throw new ChatException(ChatException.E_SOCKET_SETUP_FAILED, e);
        }

        itsThreadGroup = new ThreadGroup( THREAD_GROUP_NAME );
        itsConnections = new Vector();

        // Init DB Driver
        try {
            itsDBDriver   = theProps.getProperty(Config.PROP_DB_DRIVER);
            itsDBURL      = theProps.getProperty(Config.PROP_DB_URL);
            itsDBUser     = theProps.getProperty(Config.PROP_DB_USER);
            itsDBPassword = theProps.getProperty(Config.PROP_DB_PASSWORD);
            Class.forName(itsDBDriver).newInstance();
            DB = new PStore(this, itsDBURL, itsDBUser, itsDBPassword);

            // Init chat rooms
            itsChatRooms = DB.getChatRooms(this);

        } catch ( Exception e ) {
            throw new ChatException(ChatException.E_DB_INIT_FAILURE, e);
        }
    }

    /**
     * This method never returns, just keeps accepting connections
     * and creating connection objects to handle them.
     */
    public void run () {

        itsMainThread = Thread.currentThread();

        Socket anIncomingSocketConnection = null;
        ChatConnection aConnection = null;
        Thread aThread = null;
        long aLoopCounter = 0;

        log("ChatServer initialized, listening on port " + itsPort);

        try {
            while ( !isDone ) {
                try {
                    anIncomingSocketConnection = itsServerSocket.accept();
                    aConnection = createConnection(anIncomingSocketConnection);
                    itsConnections.add(aConnection);
                    aThread = new Thread(itsThreadGroup, aConnection);
                    aThread.start();

                } catch ( InterruptedIOException ie ) {
                    continue;
                    
                } catch ( Exception e ) {
                    log("Exception starting up new ChatConnection: " + Util.getStackTrace(e));
                    if ( anIncomingSocketConnection != null ) {
                        try { anIncomingSocketConnection.close(); } catch ( Exception e2 ) {}
                    }
                }
                
                Util.sleep( LOOP_DELAY_SLEEP_MS );
                // if ( ((++aLoopCounter) % LOOP_CLEANUP_COUNT) == 0 ) {
                cleanClosedConnections();
                // }
            }
        } finally {
            cleanup();
            log("halted.");
        }
    }

    /**
     * Called by the garbage collector just before this object is reclaimed.
     * Just calls the cleanup method.
     */
    public void finalize () {
        cleanup();
    }

    /**
     * Log some data
     * @param s the data to be logged
     */
    public synchronized void log ( String s ) {
        System.out.println( (new Date()).toString() + ": [ChatServer] " + s );
    }

    /**
     * Cleans up all open connections.  This should only be
     * called when the server is shutting down.
     */
    private void cleanup () {
        // Close server socket
        try { itsServerSocket.close(); } catch ( Exception e ) {}
        
        // Close all open connections
        if ( itsConnections != null ) {
            Iterator anIterator = itsConnections.iterator();
            ChatConnection aConnection = null;
            while ( anIterator.hasNext() ) {
                aConnection = (ChatConnection) anIterator.next();
                if ( aConnection != null && aConnection.isAlive() ) {
                    aConnection.close();
                }
            }
        }
    }

    /**
     * Cleans up connections that have completed their conversations.
     * This may be called periodically during the server's lifetime.
     */
    private void cleanClosedConnections () {
        // Remove any connections that have completed
        if ( itsConnections != null ) {
            ChatConnection aConnection = null;
            for ( int i=0; i<itsConnections.size(); i++ ) {
                aConnection = (ChatConnection) itsConnections.get(i);
                if ( aConnection != null && !aConnection.isAlive() ) {
                    log("SHUTTING DOWN CONNECTION from cleanClosedConnections!");
                    aConnection.close();
                    itsConnections.remove(i);
                    i--;
                }
            }
        }
    }

    /**
     * Creates connection processing thread for the new incoming
     * connection.
     *
     * @param theIncomingSocketConnection the socket that we need to set up a connection for
     * @return a new connection
     */ 
    protected ChatConnection createConnection( Socket theIncomingSocketConnection ) 
        throws IOException {

        return new ChatConnection( this, theIncomingSocketConnection, (itsCurrentID++) );
    }

    /**
     * Shows how to invoke this class from the command line
     */
    public static void printUsage () {
        System.out.println("\nUsage:"
                           + "\n java org.opendj.server.chat.ChatServer <prop-file>"
                           + "\n"
                           + "\n where <prop-file> is a java properties file."
                           + "\n");
    }

    public PStore getPStore() {
        return DB;
    }

    public String getProperty (String thePropName) { 
        return itsProps.getProperty(thePropName);
    }

    public ChatRoom getRoom ( String theRoomName ) {
        return (ChatRoom) itsChatRooms.get(theRoomName);
    }

    public void leaveRoom ( ChatConnection theConn, String theRoomName ) {
        ChatRoom aRoom = null;

        aRoom = (ChatRoom) itsChatRooms.get(theRoomName);
        if ( aRoom == null ) {
            log("leaveRoom: no such room: '" + theRoomName + "'");
            return;
        }

        aRoom.removeChatter( theConn );        
    }
}
