/*
 * This file and everything in this package is covered under
 * the OPENdj Software License.  See the LICENSE file at the top
 * level of this package for full details.  A copy of the license
 * is also available at: http://opendj.org/docs/licenses/LICENSE.opendj
 * 
 * Copyright 2001 DAX Interactive, LLC.
 *
 * $Author: jonathan $
 * $Revision: 1.1.1.1 $
 * $Header: /home/cvspublic/opendj/opendj-chat-server/src/org/opendj/server/chat/ChatSocketReader.java,v 1.1.1.1 2001/09/29 20:21:13 jonathan Exp $
 * $Log: ChatSocketReader.java,v $
 * Revision 1.1.1.1  2001/09/29 20:21:13  jonathan
 * Imported opendj-chat-server
 *
 */
package org.opendj.server.chat;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.DataInputStream;
import java.net.Socket;
import java.net.ServerSocket;
import java.net.SocketException;
import org.opendj.util.Util;
import org.opendj.util.BetterTokenizer;

public class ChatSocketReader implements Runnable {

    private ChatSocket     itsChatSocket     = null;
    private ChatConnection itsChatConnection = null;
    private Socket         itsSocket         = null;
    private BufferedReader itsSockReader     = null;

    private volatile Thread itsMainThread = null;
    private volatile boolean isDone = false;

    private volatile boolean isReady = false;
    private ServerSocket itsServerSocket = null;

    public ChatSocketReader ( ChatConnection theConn, ChatSocket theChatSocket ) {
        itsChatConnection = theConn;
        itsChatSocket = theChatSocket;
    }

    public void run () {

        itsMainThread = Thread.currentThread();

        String anInputLine = null;
        String aCommand    = null;
        String aData       = null;
        String aTarget     = null;
        BetterTokenizer bt = null;

        try {
            itsServerSocket = new ServerSocket(0);
            isReady = true;
            itsServerSocket.setSoTimeout(10000);

            itsSocket = itsServerSocket.accept();
            // itsSocket.setSoTimeout(Protocol.SOCKET_TIMEOUT_MILLIS);
            itsSocket.shutdownOutput();
            itsSockReader = new BufferedReader( new InputStreamReader( itsSocket.getInputStream() ) );
            isDone = false;
            log("Created ChatSocketReader for remote IP " + itsSocket.getInetAddress());

            while ( !isDone ) {
                // Just wait for the client to tell us things
                try {
                    // Reset vars
                    aCommand = null; 
                    aData = null;
                    aTarget = null; 

                    anInputLine = itsSockReader.readLine();
                    log("<" + anInputLine);
                    
                    if ( anInputLine == null ) break;
                    
                    bt = new BetterTokenizer(anInputLine, ChatSocket.SOCKDATA_DELIM);
                    
                    if ( bt.hasMoreTokens() ) {
                        aCommand = bt.nextToken();
                    }
                    if ( bt.hasMoreTokens() ) {
                        aData = bt.nextToken();
                    }
                    if ( bt.hasMoreTokens() ) {
                        aTarget = bt.nextToken();
                    }
                    
                    processCommand( aCommand, aData, aTarget );

                } catch ( SocketException se ) {
                    log("SocketException in ChatSocketReader listen-loop", se);
                    break;
                    
                } catch ( Exception e ) {
                    log("Error in ChatSocketReader listen-loop", e);
                }
            }
        } catch ( Exception e ) {
            log("Exception in run", e);

        } finally {
            log("ChatSocketReader closing from finally...");
            isDone = true;
            try { itsServerSocket.close(); } catch ( Exception e ) {}
            try { itsSocket.close(); } catch ( Exception e ) {}
            itsChatSocket.finish();
        }
    }

    private void processCommand ( String theCommand, 
                                  String theData,
                                  String theTarget ) {

        log("processCommand(cmd='" + theCommand
            + "', data='" + theData
            + "', target='" + theTarget + "'");
        theCommand = theCommand.toUpperCase();
        if ( theCommand.equals(Protocol.REQ_PONG) ) {
            itsChatConnection.pong();

        } else if ( theCommand.equals(Protocol.REQ_HELLO) ) {
            itsChatConnection.hello(theData);
 
        } else if ( theCommand.equals(Protocol.REQ_JOIN) ) {
            itsChatConnection.join(theData);

        } else if ( theCommand.equals(Protocol.REQ_JOINASDJ) ) {
            itsChatConnection.joinAsDj(theData, theTarget);

        } else if ( theCommand.equals(Protocol.REQ_JOINASDJEP) ) {
            itsChatConnection.joinAsDjWithEncryptedPassword(theData, theTarget);

        } else if ( theCommand.equals(Protocol.REQ_BOOT) ) {
            itsChatConnection.boot(theData);

        } else if ( theCommand.equals(Protocol.REQ_MESSAGE) ) {
            itsChatConnection.message(theData, theTarget);

        } else if ( theCommand.equals(Protocol.REQ_LIST_USERS) ) {
            itsChatConnection.listusers();

        } else if ( theCommand.equals(Protocol.REQ_END) ) {
            itsChatConnection.end();

        } else {
            itsChatSocket.sendData(Protocol.RES_UNRECOGNIZED_COMMAND);
        }
    }

    public void log ( String s ) {
        itsChatSocket.log("[Reader] " + s);
    }

    public void log ( String s, Exception e ) {
        itsChatSocket.log("[Reader] " + s + ": " + e.toString() + "\n" + Util.getStackTrace(e));
    }

    public void finish () {
        isDone = true;
        if ( itsMainThread != null ) {
            itsMainThread.interrupt();
        }
    }

    public boolean getIsReady() { return isReady; }

    public int getServerPort() { 
        if ( itsServerSocket != null ) {
            return itsServerSocket.getLocalPort(); 
        }
        return -1;
    }

    public boolean getIsDone() { return isDone; }
}
